import torch

from .layers import FeaturesLinear, FactorizationMachine, AnovaKernel, FeaturesEmbedding


class HighOrderFactorizationMachineModel(torch.nn.Module):
    """
    A pytorch implementation of Higher-Order Factorization Machines.

    Reference:
        M Blondel, et al. Higher-Order Factorization Machines, 2016.
    """

    def __init__(self, field_dims, order, embed_dim):
        super().__init__()
        if order < 1:
            raise ValueError(f'invalid order: {order}')
        self.order = order
        self.embed_dim = embed_dim
        self.linear = FeaturesLinear(field_dims)
        if order >= 2:
            self.embedding = FeaturesEmbedding(field_dims, embed_dim * (order - 1))
            self.fm = FactorizationMachine(reduce_sum=True)
        if order >= 3:
            self.kernels = torch.nn.ModuleList([
                AnovaKernel(order=i, reduce_sum=True) for i in range(3, order + 1)
            ])

    def forward(self, x):
        """
        :param x: Long tensor of size ``(batch_size, num_fields)``
        """
        y = self.linear(x).squeeze(1)
        if self.order >= 2:
            x = self.embedding(x)
            x_part = x[:, :, :self.embed_dim]
            y += self.fm(x_part).squeeze(1)
            for i in range(self.order - 2):
                x_part = x[:, :, (i + 1) * self.embed_dim: (i + 2) * self.embed_dim]
                y += self.kernels[i](x_part).squeeze(1)
        return torch.sigmoid(y)




import ray

from automlkiller.models.model_factory import ModelFactory
from automlkiller.utils import LOGGER


class AUTORL(object):
    def __init__(self,
                 config,
                 env,
                 env_config,
                 ):
        ray.init()
        self.config = config
        self.env = env
        self.env_config = env_config
        self.estimator = {}
        self.model = {}
        self.metrics = {}
        self.estimator_params = {}

    def create_model(self,
                     estimator,
                     n_jobs  = -1,
                     estimator_params = {}
                     ):

        estimator_model = {}
        if estimator is None:
            if len(self.estimator.keys()) > 0:
                for name_model, estimator in self.estimator.items():
                    if name_model in estimator_params.keys():
                        estimator_model[name_model] = ModelFactory.create_executor(name_model,
                                                                                   env = self.env,
                                                                                   env_config = self.env_config,
                                                                                   config = self.config,
                                                                                   **estimator_params[name_model])
                    else:
                        estimator_model[name_model] = estimator

            else:
                for name_model in ModelFactory.name_registry:
                    if name_model in estimator_params.keys():
                        estimator_model[name_model] = ModelFactory.create_executor(name_model,
                                                                                   env = self.env,
                                                                                   env_config = self.env_config,
                                                                                   config = self.config,
                                                                                   **estimator_params[name_model])
                    else:
                        estimator_model[name_model] = ModelFactory.create_executor(name_model,
                                                                                   env = self.env,
                                                                                   env_config = self.env_config,
                                                                                   config = self.config,
                                                                                   )
        else:
            for name_model in estimator:
                if name_model in estimator_params.keys():
                    estimator_model[name_model] = ModelFactory.create_executor(name_model,
                                                                               env = self.env,
                                                                               env_config = self.env_config,
                                                                               config = self.config,
                                                                               **estimator_params[name_model])
                else:
                    estimator_model[name_model] = ModelFactory.create_executor(name_model,
                                                                               env = self.env,
                                                                               env_config = self.env_config,
                                                                               config = self.config,
                                                                               )

        # update estimator_params
        for name_model, params in estimator_params.items():
            self.estimator_params[name_model] = params

        for name_model, model in estimator_model.items():
            print('name_model: ', name_model)
            try:
                estimator = model.estimator
            except:
                estimator = model
            estimator.fit(*estimator_params)
            results = estimator.predict()
            name_model = ''.join(name_model.split('-')[1:])

            self.metrics[name_model] = results['reward']
        return self
    def report_rl(self):
        print('report: ', self.metrics)






from .a2c import A2CContainer
from .a3c import A3CContainer
from .appo import APPOContainer
# from .bc import BCContainer
from .ddpg import DDPGContainer
from .ddppo import DDPPOContainer
from .dqn import DQNContainer
# from .dreamer import DREAMERContainer
from .es import ESContainer
# from .impala import IMPALAContainer
from .maml import MAMLContainer
from .marwil import MARWILContainer
from .mbmpo import MBMPOContainer
from .pg import PGContainer
from .ppo import PPOContainer
from .sac import SACContainer
from .td3 import TD3Container






from .affinitypropagation_cluster import AffinityPropagation
from .agglomerativeclustering_cluster import AgglomerativeClustering
from .birch_cluster import BirchContainer
from .dbscan_cluster import DBSCANContainer
from .kmeans_cluster import KMeansContainer
from .kmodes_cluster import KModesContainer
from .meanshift_cluster import MeanShiftContainer
from .optics_cluster import OPTICSContainer
from .spectral_cluster import SpectralClusteringContainer


from sklearn.linear_model import Lasso
from automlkiller.models.model_factory import ModelFactory
from automlkiller.utils.distributions import np_list_arange, UniformDistribution, IntUniformDistribution

@ModelFactory.register('regression-lasso')
class LassoRegressionContainer(Lasso):
    def __init__(self, **kwargs):
        super().__init__()
        tune_grid = {}
        tune_distributions = {}

        tune_grid = {
            "alpha": np_list_arange(0.01, 10, 0.01, inclusive=True),
            "fit_intercept": [True, False],
            "normalize": [True, False],
        }
        tune_distributions = {"alpha": UniformDistribution(0.001, 10)}

        self.tune_grid = tune_grid
        self.tune_distributions = tune_distributions
        self.estimator = Lasso(**kwargs)



from enum import Enum, auto
import numpy as np
import pandas as pd
import sklearn
from imblearn.pipeline import Pipeline
from automlkiller.preprocess.preprocess import *
from automlkiller.preprocess.preprocee_factory import PreprocessFactory
from automlkiller.utils import LOGGER

class Preprocess(object):
    def __init__(self, **kwargs):
        super(Preprocess, self).__init__()
        self.preprocess_objs = []
        for name_registry, preprocess_param in kwargs.items():
            name_registry = '-'.join(['preprocess', name_registry])
            if name_registry not in PreprocessFactory.registry.keys():
                LOGGER.info(f'{name_registry} not PreprocessFactory')
            LOGGER.info('preprocess registry {}'.format(name_registry))
            preprocess_object = PreprocessFactory.create_executor(name = name_registry,
                                                                  **preprocess_param)
            self.preprocess_objs.append(preprocess_object)

    def fit(self, X, y = None, **fit_params):
        X = X.copy()
        if y is not None:
            y = y.copy()
            if isinstance(y, pd.Series):
                y = pd.DataFrame(y)
            y.columns = [str(col) for col in y.columns.tolist()]
        if isinstance(X, pd.Series):
            X = pd.DataFrame(X)
        X.columns = [str(col) for col in X.columns.tolist()]
        X = X.loc[:, ~X.columns.duplicated()]
        X.dropna(axis=1, how='all', inplace=True)
        for obj in self.preprocess_objs:
            obj.fit(X, y)

    def transform(self, X, y = None, **fit_params):
        X = X.copy()
        if y is not None:
            y = y.copy()
            if isinstance(y, pd.Series):
                y = pd.DataFrame(y)
            y.columns = [str(col) for col in y.columns.tolist()]

        if isinstance(X, pd.Series):
            X = pd.DataFrame(X)
        X.columns = [str(col) for col in X.columns.tolist()]
        X = X.loc[:, ~X.columns.duplicated()]
        X.dropna(axis=1, how='all', inplace=True)
        data = X, y
        for obj in self.preprocess_objs:
            data = obj.transform(*data)
        if isinstance(data, pd.DataFrame) ==False:
            return data[0]
        return data
    def fit_transform(self, X, y = None):
        for obj in self.preprocess_objs:
            obj.fit(X, y)
            X, y = obj.transform(X, y)

        if y is None:
            return X
        return X, y







import os
import numpy as np
import torch
import torch.nn as nn
import pytorch_lightning as pl

from torch.utils.data import DataLoader
from sklearn.datasets import make_classification
from layers.layer import FactorizationMachine, FeaturesEmbedding, FeaturesLinear, MultiLayerPerceptron

class DeepFactorizationMachineModel(pl.LightningModule):
    def __init__(self, field_dims, embed_dim, mlp_dims, dropout):
        super().__init__()
        self.linear = FeaturesLinear(field_dims)
        self.fm = FactorizationMachine(reduce_sum = True)
        self.embedding = FeaturesEmbedding(field_dims, embed_dim)
        self.embed_output_dim = len(field_dims) *embed_dim
        self.mlp = MultiLayerPerceptron(self.embed_output_dim, mlp_dims, dropout)


    def forward(self, x):
        embed_x = self.embedding(x)
        x = self.linear(x) + self.fm(embed_x) + self.mlp(embed_x.view(-1, self.embed_output_dim))
        return torch.sigmoid(x.squeeze(1))

class FMDataset(torch.utils.data.Dataset):
    def __init__(self):
        super().__init__()
        self.items = np.random.randint(10, size=(1000, 20))
        self.targets = np.random.randn(1000, 10)
        self.field_dims = np.max(self.items, axis = 0) + 1

    def __len__(self):
        return 1000
    def __getitem__(self, index):
        return torch.from_numpy(self.items[index])
if __name__ == '__main__':
    dataset = FMDataset()
    field_dims = dataset.field_dims
    loader = DataLoader(dataset, batch_size = 4, num_workers=8)
    model = DeepFactorizationMachineModel(field_dims = [10],
                                      embed_dim = 40)
    for field in loader:
        out = model(field)
        print('out: ', out)
        print('shape: ', out.shape)
        break
